# -*- coding: UTF-8 -*-
# @Time : 2022/4/12 下午2:45 
# @Author : 刘洪波
"""
schema 消息类型
参考网址：https://pulsar.apache.org/docs/en/client-libraries-python/
Schema	Notes
BytesSchema	Get the raw payload as a bytes object. No serialization/deserialization are performed. This is the default schema mode
StringSchema	Encode/decode payload as a UTF-8 string. Uses str objects
JsonSchema	Require record definition. Serializes the record into standard JSON payload
AvroSchema	Require record definition. Serializes in AVRO format
"""
import pulsar


def StringSchema():
    return pulsar.schema.StringSchema()


def BytesSchema():
    return pulsar.schema.BytesSchema()


def JsonSchema(record_cls):
    return pulsar.schema.JsonSchema(record_cls)


def AvroSchema(record_cls):
    return pulsar.schema.AvroSchema(record_cls)



# -*- coding: UTF-8 -*-
# @Time : 2022/5/6 下午4:49 
# @Author : 刘洪波
from retry import retry
import pika
from pika import exceptions
from concurrent.futures import ThreadPoolExecutor
import loguru
import time
"""
消费者
"""


class Consumer(object):
    def __init__(self, host, port, username, password, heartbeat, exchange, routing_key, durable=False):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.heartbeat = heartbeat
        self.exchange = exchange
        self.routing_key = routing_key
        self.durable = durable
        self.cosumed_num = 0  # 已经消费的次数

    @retry(pika.exceptions.AMQPConnectionError, delay=5, jitter=(1, 3), logger=loguru.logger)
    def receive(self, task, thread_count=None, consume_num=0):
        """
        消费者消费
        :param task:
        :param thread_count: 当thread_count 有值的时候，可以进行多线程并行消费
        :param consume_num: 消费consume_num 次就停止消费
        :return:
        """
        if thread_count is not None:
            if not isinstance(thread_count, int):
                raise ValueError('error type of parameter: message_list must be of type int')
        if not isinstance(consume_num, int):
            raise ValueError('error type of parameter: message_list must be of type int')
        loguru.logger.info('开始连接rabbitmq')
        connection = pika.BlockingConnection(pika.ConnectionParameters(
            host=self.host, port=self.port, credentials=pika.PlainCredentials(self.username, self.password),
            heartbeat=self.heartbeat)
        )
        loguru.logger.info('已连接至rabbitmq')
        channel = connection.channel()
        channel.exchange_declare(exchange=self.exchange, exchange_type='topic', durable=self.durable)
        if consume_num:
            queue_name = self.routing_key + '.' + str(int(round(time.time() * 1000000)))
        else:
            queue_name = self.routing_key
        try:
            channel.queue_declare(queue=queue_name, passive=True)
        except Exception as e:
            channel = connection.channel()
            channel.exchange_declare(exchange=self.exchange, exchange_type='topic', durable=self.durable)
            channel.queue_declare(queue=queue_name)
            channel.queue_bind(exchange=self.exchange, queue=queue_name, routing_key=self.routing_key)

        if thread_count:
            pool = ThreadPoolExecutor(max_workers=thread_count)

        def callback(ch, method, properties, body):
            body = body.decode('utf8')
            if thread_count:
                pool.submit(task, body)
            else:
                task(body)
            ch.basic_ack(delivery_tag=method.delivery_tag)
            if consume_num:
                self.cosumed_num += 1
                if self.cosumed_num == consume_num:
                    ch.queue_delete(queue=queue_name)
                    ch.stop_consuming()

        channel.basic_consume(queue=queue_name, auto_ack=False, on_message_callback=callback)
        channel.basic_qos(prefetch_count=1)
        channel.start_consuming()



# -*- coding: UTF-8 -*-
# @Time : 2021/11/28 上午12:03 
# @Author : 刘洪波
import pulsar


class Client(object):
    def __init__(self, url: str):
        self.client = pulsar.Client(url)

    def create_consumer(self, topic, consumer_name: str, schema=pulsar.schema.StringSchema(), consumer_type='Shared'):
        """
        创建 消费者
        :param topic
        :param consumer_name: 消费者名字
        :param schema:
        :param consumer_type:
        :return:
        """
        from pulsar_mq.Consumers import Consumer
        return Consumer(self.client, topic, consumer_name, schema, consumer_type)

    def create_producer(self, topic: str, schema=pulsar.schema.StringSchema()):
        """
        创建生产者
        :param topic: topic
        :param schema:
        :return:
        """
        from pulsar_mq.Producers import Producer
        return Producer(self.client, topic, schema)

    def consume_produce(self, consumer_topic, consumer_name: str, producer_topic=None,
                        schema=pulsar.schema.StringSchema(), consumer_type='Shared'):
        """
        服务端
        pulsar 消费数据后 并且发送数据的 服务
        1. 订阅pulsar
        2. 处理消费的数据
        3. 发送得到的结果
        4. 发送的 队列 可以是 随机的
        :param consumer_topic:
        :param consumer_name:
        :param producer_topic:
        :param schema:
        :param consumer_type:
        :return:
        """
        from pulsar_mq.Services import ConsumeProduce
        return ConsumeProduce(self.client, consumer_topic, consumer_name, consumer_type, producer_topic, schema)

    def produce_consume(self, producer_topic: str, consumer_topic=None, consumer_name=None,
                        schema=pulsar.schema.StringSchema(), consumer_type='Shared'):
        """
        调用端
        发送数据至pulsar，然后从 pulsar消费数据的服务
        1. 发送数据至pulsar
        2. 订阅pulsar得到业务数据
        3. 消费的 队列 可以是 随机的
        :param producer_topic:
        :param consumer_topic:
        :param consumer_name:
        :param schema:
        :param consumer_type:
        :return:
        """
        from pulsar_mq.Services import ProduceConsume
        return ProduceConsume(self.client, producer_topic, consumer_type, consumer_topic, consumer_name, schema)

    def close(self):
        """
        关闭 client
        :return:
        """
        self.client.close()



# -*- coding: UTF-8 -*-
# @Time : 2021/12/3 下午7:05 
# @Author : 刘洪波
import time
from concurrent.futures import ThreadPoolExecutor
from pulsar_mq.Clients import Client
from rabbitmq.Connections import Connection


class RabbitmqToPulsar(object):
    def __init__(self, rabbitmq_connect: Connection, pulsar_client: Client):
        """
        rabbitmq 和 pulsar 连接
        """
        self.rabbitmq_connect = rabbitmq_connect
        self.pulsar_client = pulsar_client

    def run(self, task, pulsar_topic, rabbitmq_exchange, rabbitmq_routing_key, durable=False,
            rabbitmq_thread_count=None, _async=True, send_callback=None, random_topic=None, logger=None):
        """
        1. 订阅rabbitmq
        2. 处理消费的数据
        3. 将处理后的数据 发送到 pulsar
        :param task:  该函数是处理消费的数据
        :param pulsar_topic:
        :param random_topic: 随机 队列
        :param rabbitmq_exchange:
        :param rabbitmq_routing_key:
        :param durable:
        :param rabbitmq_thread_count:
        :param _async:  pulsar是否异步发送
        :param send_callback: pulsar异步发送时的回调函数
        :param logger: 日志收集器
        :return:
        """
        try:
            producer = self.pulsar_client.create_producer(pulsar_topic)
            consumer = self.rabbitmq_connect.create_consumer(rabbitmq_exchange, rabbitmq_routing_key, durable)

            def callback(msg):
                result = task(msg)
                if result:
                    producer.send(result, _async=_async, callback=send_callback, random_topic=random_topic)

            consumer.receive(callback, rabbitmq_thread_count)
        except Exception as e:
            if logger:
                logger.error(e)
            else:
                print(e)


class PulsarToRabbitmq(object):
    def __init__(self, rabbitmq_connect: Connection, pulsar_client: Client):
        """
        rabbitmq 和 pulsar 连接
        """
        self.rabbitmq_connect = rabbitmq_connect
        self.pulsar_client = pulsar_client

    def run(self, task, rabbitmq_exchange, rabbitmq_routing_key,
            pulsar_consumer_topic=None, pulsar_consumer_name=None,
            durable=False, pulsar_thread_count=None, logger=None):
        """
        1. 订阅 pulsar
        2. 处理消费的数据
        3. 将处理后的数据发送到 rabbitmq
        :param task:
        :param pulsar_consumer_topic:
        :param pulsar_consumer_name:
        :param rabbitmq_exchange:
        :param rabbitmq_routing_key:
        :param pulsar_thread_count:
        :param durable:
        :param logger:
        :return:
        """
        try:
            if pulsar_consumer_topic and pulsar_consumer_name:
                consumer = self.pulsar_client.create_consumer(pulsar_consumer_topic, pulsar_consumer_name)
            else:
                raise ValueError('pulsar error: consumer_topic and random_topic is None, need consumer topic')

            producer = self.rabbitmq_connect.create_producer(rabbitmq_exchange, rabbitmq_routing_key, durable)

            def callback(msg):
                result = task(msg)
                if result:
                    producer.send(result)

            consumer.receive(callback, pulsar_thread_count, logger)
        except Exception as e:
            if logger:
                logger.error(e)
            else:
                print(e)


class InterService(object):
    def __init__(self, rabbitmq_connect, pulsar_client):
        """
        rabbitmq 和 pulsar 连接
        """
        self.rabbitmq_connect = rabbitmq_connect
        self.pulsar_client = pulsar_client
        self.rabbitmq_to_pulsar = RabbitmqToPulsar(rabbitmq_connect, pulsar_client)
        self.pulsar_to_rabbitmq = PulsarToRabbitmq(rabbitmq_connect, pulsar_client)

    def run(self, pulsar_producer_topic, rabbitmq_send_exchange, rabbitmq_send_routing_key,
            rabbitmq_consumer_exchange,
            rabbitmq_consumer_routing_key, pulsar_consumer_topic, pulsar_consumer_name, send_task=None,
            consumer_task=None, durable=False,
            consumer_durable=None, producer_durable=None,
            _async=True, send_callback=None, rabbitmq_thread_count=None, pulsar_thread_count=None,
            logger=None):
        """
        1. 从 rabbitmq 订阅，将数据发送至 pulsar;
        2. 从 pulsar 订阅，将数据发送至 rabbitmq
        :param send_task:
        :param consumer_task:
        :param pulsar_producer_topic:
        :param pulsar_consumer_topic:
        :param pulsar_consumer_name:
        :param rabbitmq_send_exchange:
        :param rabbitmq_send_routing_key:
        :param rabbitmq_consumer_exchange:
        :param rabbitmq_consumer_routing_key:
        :param durable:
        :param consumer_durable:
        :param producer_durable:
        :param _async:
        :param send_callback:
        :param rabbitmq_thread_count:
        :param pulsar_thread_count:
        :param logger:
        :return:
        """
        if logger:
            def send_task2(msg):
                logger.info('取rabbitmq的消息：')
                logger.info(msg)
                return msg

            def consumer_task2(msg):
                logger.info('取pulsar的消息：')
                logger.info(msg.data())
                return [msg.data()]
        else:
            def send_task2(msg):
                return msg

            def consumer_task2(msg):
                return [msg.data()]

        if send_task is None:
            send_task = send_task2
        if consumer_task is None:
            consumer_task = consumer_task2
        pool = ThreadPoolExecutor(max_workers=2)
        pool.submit(self.rabbitmq_to_pulsar.run, send_task, pulsar_producer_topic, rabbitmq_consumer_exchange,
                    rabbitmq_consumer_routing_key, durable=consumer_durable if consumer_durable else durable,
                    rabbitmq_thread_count=rabbitmq_thread_count,
                    _async=_async, send_callback=send_callback, logger=logger)
        pool.submit(self.pulsar_to_rabbitmq.run, consumer_task, rabbitmq_send_exchange, rabbitmq_send_routing_key,
                    pulsar_consumer_topic=pulsar_consumer_topic, pulsar_consumer_name=pulsar_consumer_name,
                    durable=producer_durable if producer_durable else durable,
                    pulsar_thread_count=pulsar_thread_count, logger=logger)


class InterServiceRQ(object):
    def __init__(self, rabbitmq_connect: Connection, pulsar_client: Client):
        """
        rabbitmq 和 pulsar 连接
        """
        self.rabbitmq_connect = rabbitmq_connect
        self.pulsar_client = pulsar_client

    def run(self, pulsar_producer_topic, rabbitmq_send_exchange, rabbitmq_send_routing_key,
            rabbitmq_consumer_exchange, rabbitmq_consumer_routing_key,
            send_task=None, consumer_task=None, durable=False,
            consumer_durable=None, producer_durable=None,
            _async=True, send_callback=None, thread_count=5, logger=None):
        """
        从 rabbitmq 订阅，将数据发送至 pulsar;
        再从 pulsar 订阅，将数据发送至 rabbitmq
        :param send_task:
        :param consumer_task:
        :param pulsar_producer_topic:
        :param rabbitmq_send_exchange:
        :param rabbitmq_send_routing_key:
        :param rabbitmq_consumer_exchange:
        :param rabbitmq_consumer_routing_key:
        :param durable:
        :param consumer_durable:
        :param producer_durable:
        :param _async:
        :param send_callback:
        :param thread_count:
        :param logger:
        :return:
        """
        if logger:
            def send_task2(msg):
                logger.info('rabbitmq的输入：')
                logger.info(msg)
                return msg

            def consumer_task2(msg):
                logger.info('pulsar的输出：')
                logger.info(msg.data())
                return [msg.data()]
        else:
            def send_task2(msg):
                return msg

            def consumer_task2(msg):
                return [msg.data()]

        if send_task is None:
            send_task = send_task2
        if consumer_task is None:
            consumer_task = consumer_task2

        rb_consumer = self.rabbitmq_connect.create_consumer(rabbitmq_consumer_exchange, rabbitmq_consumer_routing_key,
                                                            consumer_durable if consumer_durable else durable)
        rb_producer = self.rabbitmq_connect.create_producer(rabbitmq_send_exchange, rabbitmq_send_routing_key,
                                                            producer_durable if producer_durable else durable)

        producer = self.pulsar_client.create_producer(pulsar_producer_topic)
        pool = ThreadPoolExecutor(max_workers=thread_count)

        def callback(msg):
            result = send_task(msg)
            if result:
                random_topic = 'random_topic_' + str(int(round(time.time() * 1000000)))

                def send_consume():
                    consumer = self.pulsar_client.create_consumer(random_topic, random_topic)
                    producer.send(result, _async=_async, callback=send_callback, random_topic=random_topic)

                    def callback_task(_msg):
                        consumer_result = consumer_task(_msg)
                        if consumer_result:
                            rb_producer.send(consumer_result)

                    consumer.receive_one(callback_task, logger)

                pool.submit(send_consume)

        rb_consumer.receive(callback)



# -*- coding: UTF-8 -*-
# @Time : 2022/5/6 下午4:51 
# @Author : 刘洪波
from rabbitmq.Consumers import Consumer
from rabbitmq.Producers import Producer


class ConsumeProduce(object):
    def __init__(self, consumer: Consumer, producer: Producer):
        self.consumer = consumer
        self.producer = producer

    def run(self, task,  thread_count=None):
        """
        rabbitmq 服务端
        1. 订阅rabbitmq
        2. 处理消费的数据
        3. 发送得到的结果
        :param task:
        :param thread_count:
        :return:
        """

        def callback(body):
            result = task(body)
            if result:
                self.producer.send(result)

        self.consumer.receive(callback, thread_count)


class ProduceConsume(object):
    def __init__(self, consumer: Consumer, producer: Producer):
        self.consumer = consumer
        self.producer = producer
        self.result = []

    def run(self, message_list: list, thread_count=None):
        """
        rabbitmq 调用端
        1. 往发rabbitmq送消息
        2. 订阅rabbitmq
        3. 消费得到服务端返回的结果
        :param message_list:
        :param thread_count:
        :return:
        """
        if message_list:
            self.producer.send(message_list)

            def callback(body):
                self.result.append(body)

            self.consumer.receive(callback, thread_count, len(message_list))
        return self.result



# -*- coding: UTF-8 -*-
# @Time : 2021/12/3 下午6:36 
# @Author : 刘洪波


def connect(host, port, username, password, url, heartbeat=60):
    from RabbitmqPulsar.Connections import Interconnection
    return Interconnection(host, port, username, password, url, heartbeat)



# -*- coding: UTF-8 -*-
# @Time : 2021/12/2 下午5:55
# @Author : 刘洪波
from rabbitmq.Consumers import Consumer
from rabbitmq.Producers import Producer
from rabbitmq.Services import ConsumeProduce, ProduceConsume


class Connection(object):
    def __init__(self, host, port, username, password, heartbeat):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.heartbeat = heartbeat

    def create_consumer(self, exchange, routing_key, durable=False):
        """
        消费者
        :param exchange:
        :param routing_key:
        :param durable:
        :return:
        """
        return Consumer(self.host, self.port, self.username, self.password,
                        self.heartbeat, exchange, routing_key, durable)

    def create_producer(self, exchange, routing_key, durable=False):
        """
        生产者
        :param exchange:
        :param routing_key:
        :param durable:
        :return:
        """
        return Producer(self.host, self.port, self.username, self.password,
                        self.heartbeat, exchange, routing_key, durable)

    def consume_produce(self, consumer_exchange, consumer_routing_key,
                        producer_exchange, producer_routing_key, durable=False):
        """
        服务端（先消费再生产）
        :param consumer_exchange:
        :param consumer_routing_key:
        :param producer_exchange:
        :param producer_routing_key:
        :param durable:
        :return:
        """
        return ConsumeProduce(self.create_consumer(consumer_exchange, consumer_routing_key, durable),
                              self.create_producer(producer_exchange, producer_routing_key, durable))

    def produce_consume(self, consumer_exchange, consumer_routing_key,
                        producer_exchange, producer_routing_key, durable=False):
        """
        调用端（先生产再消费）
        :param consumer_exchange:
        :param consumer_routing_key:
        :param producer_exchange:
        :param producer_routing_key:
        :param durable:
        :return:
        """
        return ProduceConsume(self.create_consumer(consumer_exchange, consumer_routing_key, durable),
                              self.create_producer(producer_exchange, producer_routing_key, durable))



# -*- coding: UTF-8 -*-
# @Time : 2022/4/15 下午9:51 
# @Author : 刘洪波
import pulsar_mq
import rabbitmq


"""
rabbitmq 和 pulsar互相订阅消费
"""


class Interconnection(object):
    def __init__(self, rabbitmq_host, rabbitmq_port, rabbitmq_username, rabbitmq_password, pulsar_url, heartbeat):
        """
        rabbitmq 和 pulsar 连接
        """
        self.rabbitmq_connect = rabbitmq.connect(rabbitmq_host, rabbitmq_port,
                                                 rabbitmq_username, rabbitmq_password, heartbeat)
        self.pulsar_client = pulsar_mq.client(pulsar_url)

    def rabbitmq_to_pulsar(self):
        """
        1. 订阅rabbitmq
        2. 处理消费的数据
        3. 将处理后的数据 发送到 pulsar
        """
        from RabbitmqPulsar.Services import RabbitmqToPulsar
        return RabbitmqToPulsar(self.rabbitmq_connect, self.pulsar_client)

    def pulsar_to_rabbitmq(self):
        """
        1. 订阅 pulsar
        2. 处理消费的数据
        3. 将处理后的数据发送到 rabbitmq
        """
        from RabbitmqPulsar.Services import PulsarToRabbitmq
        return PulsarToRabbitmq(self.rabbitmq_connect, self.pulsar_client)

    def inter_services(self, start_with_rabbitmq=True, random_queue=True):
        """
        rabbitmq 与 pulsar 互联服务

        :param start_with_rabbitmq:
                1. 值为True时， 从 rabbitmq 订阅，将数据发送至 pulsar;
                   并且从 pulsar 订阅，将数据发送至 rabbitmq

                2. 值为False时，从 pulsar 订阅，将数据发送至 rabbitmq；
                   并且从 rabbitmq 订阅，将数据发送至 pulsar;
        :param random_queue:  当start_with_rabbitmq=True时，pulsar是否使用随机队列来生产消息，
                              当start_with_rabbitmq=False时, 此参数不工作

                1. 值为True时，pulsar使用随机队列来生产消息
                2. 值为False时，pulsar不使用随机队列来生产消息

                建议使用随机队列
        :return:
        """
        if start_with_rabbitmq:
            if random_queue:
                from RabbitmqPulsar.Services import InterServiceRQ
                return InterServiceRQ(self.rabbitmq_connect, self.pulsar_client)
            from RabbitmqPulsar.Services import InterService
            return InterService(self.rabbitmq_connect, self.pulsar_client)
        else:
            from RabbitmqPulsar.Services import InterService
            return InterService(self.rabbitmq_connect, self.pulsar_client)



# -*- coding: UTF-8 -*-
# @Time : 2022/5/6 下午4:50 
# @Author : 刘洪波
import pika
"""
生产者
生产数据
"""


class Producer(object):
    def __init__(self, host, port, username, password, heartbeat, exchange, routing_key, durable=False):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.exchange = exchange
        self.heartbeat = heartbeat
        self.routing_key = routing_key
        self.durable = durable

    def send(self, message_list: list):
        """
        发送数据
        :param message_list: 可以一次发送多条消息
        :return:
        """
        if not isinstance(message_list, list):
            raise ValueError('error type of parameter: message_list must be of type list')
        connection = pika.BlockingConnection(pika.ConnectionParameters(
            host=self.host, port=self.port, credentials=pika.PlainCredentials(self.username, self.password),
            heartbeat=self.heartbeat)
        )
        channel = connection.channel()
        channel.exchange_declare(exchange=self.exchange, exchange_type='topic', durable=self.durable)
        for message in message_list:
            channel.basic_publish(exchange=self.exchange, routing_key=self.routing_key, body=message)
        connection.close()



# -*- coding: UTF-8 -*-
# @Time : 2021/12/2 下午5:54 
# @Author : 刘洪波
from rabbitmq.Connections import Connection


def connect(host, port, username, password, heartbeat=60):
    # heartbeat 心跳检测，默认60秒检测一次心跳
    return Connection(host, port, username, password, heartbeat)
